/*  Copyright (C) 2015 Rainer Wichmann                                     */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include "config_xor.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <ctype.h>
#include <limits.h>
#include <errno.h>

#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#define NO_DB_UPGRADE 1

/* #define SYSLOG_DBG */ 
#undef SYSLOG_DBG 

#ifdef SYSLOG_DBG
#include <syslog.h>
#define SYSLOG_MODE LOG_USER|LOG_INFO
#endif

#ifdef MAJOR_IN_MKDEV
#include <sys/mkdev.h>
#else
#ifdef MAJOR_IN_SYSMACROS
#include <sys/sysmacros.h>
#endif
#endif

#include <pwd.h>


static FILE * f_tmp = NULL;

typedef enum { read_replace, read_merge } rFlag;

static int sh_getopt_get (int argc, char * argv[]);

static int    g_verbose = 0;

static int    g_update  = 0;



static char * g_database = NULL;
static char * g_merge    = NULL;


static char * o_database = NULL;


#define QUOTE_CHAR '='
#if 0
#define SH_BUFSIZE     1024
#endif
#define SH_BUFSIZE     32767
#define MAX_PATH_STORE 12287

#define KEY_LEN   48
#define USER_MAX  20
#define GROUP_MAX 20

#undef  S_TRUE
#define S_TRUE    1
#undef  S_FALSE
#define S_FALSE   0

#define SH_DEADFILE 0x44454144

/* An unsigned integer guaranteed to be 32 bit.
 */
#if defined(HAVE_INT_32)
#define UINT32 unsigned int
#define SINT32 int
#elif defined(HAVE_LONG_32)
#define UINT32 unsigned long
#define SINT32 long
#elif defined(HAVE_SHORT_32)
#define UINT32 unsigned short
#define SINT32 short
#endif

#ifdef HAVE_INTTYPES_H
#include <inttypes.h>
#endif
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif

#if !defined(HAVE_INTTYPES_H) && !defined(HAVE_STDINT_H)

#ifdef HAVE_LONG_64
#define  UINT64 unsigned long
#defi
#else
#ifdef HAVE_LONG_LONG
#define  UINT64 unsigned long long
#else
#define UINT64_IS_32
#define  UINT64 unsigned long
#endif
#endif

#else
#define  UINT64 uint64_t
#endif


static char * program_name  = NULL;

static int  entry_undo   = 0;

static int         ISNEW_CURR   = 1;
static int         ISNEW_MERGE  = 1;

#ifndef SIZE_MAX
#define SIZE_MAX              (4294967295U)
#endif


static /*@noreturn@*/ void die (int line, const char *format, ...)
{
  va_list ap;

  if (program_name != NULL)
    fprintf (stderr, "Error: %s: line %d: uid %d: euid %d: errno %d: ", 
	     program_name, line, (int) getuid(), (int) geteuid(), errno);
  else
    fprintf (stderr, "Error: line %d: uid %d: euid %d: errno %d: ", 
	     line, (int) getuid(), (int) geteuid(), errno);

  va_start (ap, format);
  (void) vfprintf (stderr, format, ap);
  va_end (ap);

  exit(EXIT_FAILURE);
}

#define CHECK_ERR(a)				\
do { \
  if ((a) <= 0) die(__LINE__, "Error writing to file\n");	\
} while (1 == 0)

static void pverb (int verbosity, int line, const char *format, ...)
{
  va_list ap;

  if (g_verbose < verbosity)
    return;

  if (program_name != NULL)
    fprintf (stdout, "%s: %4d: ", program_name, line);
  else
    fprintf (stdout, "%4d: ", line);
  va_start (ap, format);
  (void) vfprintf (stdout, format, ap);
  va_end (ap);
  return;
}

#ifdef SH_STEALTH
#ifndef SH_MAX_GLOBS
#define SH_MAX_GLOBS 16
#endif

#ifndef GLOB_LEN
#define GLOB_LEN 255
#endif

char * globber(char * str)
{
  int i;
  int j;

  static   int  items = 0;
  static   int  count = 0;
  static   char glob[SH_MAX_GLOBS * (GLOB_LEN+1)];

  if (str == NULL)
    return NULL;
  else
    j = strlen(str);

  ++items;

  if (j > GLOB_LEN) 
    die (__LINE__, _("j <= GLOB_LEN\n"));

  if (j > GLOB_LEN) 
    j = GLOB_LEN;

  /* Overwrap the buffer.
   */
  if ( (count + j) >= (SH_MAX_GLOBS * (GLOB_LEN+1)))
    {
      count = 0;
      items = 0;
    }

  for (i = 0; i < j; ++i)
    {
      if (str[i] != '\n' && str[i] != '\t' && str[i] != '\r' && str[i] != '"')
	glob[count + i] = str[i] ^ XOR_CODE;
      else
	glob[count + i] = str[i];
    }
  glob[count + j] = '\0';

  i     = count;
  count = count + j + 1;
  return &glob[i];
}

#define sh_do_decode sh_do_encode
static void sh_do_encode (char * str, int len)
{
  register          int i;

  /* this is a symmetric operation
   */
  for (i = 0; i < len; ++i)
    {
      str[i] = str[i] ^ XOR_CODE;
    }
  return;
}

#endif

static UINT32 * swap_32 (UINT32 * iptr)
{
#ifdef WORDS_BIGENDIAN
  unsigned char swap;
  unsigned char * ii = (unsigned char *) iptr;
  swap = ii[0]; ii[0] = ii[3]; ii[3] = swap;
  swap = ii[1]; ii[1] = ii[2]; ii[2] = swap;
  return iptr;
#else
  return iptr;
#endif
}

static UINT64 *  swap_64 (UINT64 * iptr)
{
#ifdef WORDS_BIGENDIAN
#ifdef UINT64_IS_32
  swap_32 ((UINT32*) iptr);
#else
  unsigned char swap;
  unsigned char * ii = (unsigned char *) iptr;
  swap = ii[0]; ii[0] = ii[7]; ii[7] = swap;
  swap = ii[1]; ii[1] = ii[6]; ii[6] = swap;
  swap = ii[2]; ii[2] = ii[5]; ii[5] = swap;
  swap = ii[3]; ii[3] = ii[4]; ii[4] = swap;
#endif
  return iptr;
#else
  return iptr;
#endif
}

static unsigned short *  swap_short (unsigned short * iptr)
{
#ifdef WORDS_BIGENDIAN
  if (sizeof(short) == 4)
    swap_32 ((UINT32*) iptr);
  else
    {
      /* alignment problem */
      unsigned char swap;
      static unsigned short ooop;
      unsigned char * ii;
      ooop = *iptr;
      ii = (unsigned char *) &ooop;
      swap = ii[0]; ii[0] = ii[1]; ii[1] = swap;
      return &ooop;
    }
#else
  return iptr;
#endif
}

/* MAX_PATH_STORE must be >= KEY_LEN
 * is defined at top
 * #define MAX_PATH_STORE SH_BUFSIZE_1024 + 1
 */
#define ATTRBUF_SIZE 16
#define CMODE_SIZE   11

typedef struct store_info_old {
  UINT32           mode;
  UINT32           linkmode;

  UINT64           dev;
  UINT64           rdev;
  UINT32           hardlinks;
  UINT32           ino;
  UINT64           size;
  UINT64           atime;
  UINT64           mtime;
  UINT64           ctime;
  UINT32           owner;
  UINT32           group;

  UINT32           attributes;
  char             c_attributes[ATTRBUF_SIZE];

  unsigned short   mark;
  char             c_owner[USER_MAX+2];
  char             c_group[GROUP_MAX+2];
  char             c_mode[CMODE_SIZE];
  char             checksum[KEY_LEN+1];
} sh_filestore_old_t;

typedef struct store_info {

  UINT32           mode;
  UINT32           linkmode;

  UINT64           dev;
  UINT64           rdev;

  UINT32           hardlinks;
  UINT32           ino;

  UINT64           size;
  UINT64           atime;
  UINT64           mtime;
  UINT64           ctime;

  UINT32           owner;
  UINT32           group;

  UINT32           attributes;

  char             c_attributes[ATTRBUF_SIZE]; /* 16 = 2*UINT64 */

  unsigned short   mark;
  char             c_owner[USER_MAX+2];
  char             c_group[GROUP_MAX+2];
  char             c_mode[CMODE_SIZE];
  char             checksum[KEY_LEN+1];
  
  /* If 'checkflags' is elsewhere, the compiler would still use
   * a 6-byte padding to align the whole struct to an 8-byte boundary.
   * ipad, opad: make explicit what the compiler does on a 64-byte system.
   */
  char             ipad[2];
  UINT32           checkflags;
  char             opad[4];

} sh_filestore_t;

typedef struct file_info {
  sh_filestore_t   theFile;
  char           * fullpath;
  char           * linkpath;
  char           * attr_string;
  int              visited;
  int              reported;
  int              allignore;
  unsigned long    modi_mask;
  struct           file_info * next;
  UINT32           hashkey;
} sh_file_t;

/* changed again for V1.4       */
#define OLD_REC_MAGIC 21
#define REC_MAGIC 22

#define REC_FLAGS_ATTR (1<<8)
#define REC_FLAGS_MASK 0xFF00

static void * xalloc (int line, size_t size)
{
  void * ptr = malloc (size);
  if (ptr == NULL)
    die(line, "Out of memory\n");
  return ptr;
}

static char * xstrdup(int line, char * str)
{
  size_t len = strlen(str) + 1;
  char * out = xalloc(line, len);
  return strcpy(out, str);
}

/*
 * Appends src to string dst of size siz (unlike strncat, siz is the
 * full size of dst, not space left).  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 */
static int sl_strlcat(char * dst, const char *src, size_t siz)
{
  register size_t dst_end;
  register size_t dst_free;
  register size_t src_len;

  if (dst == NULL)
    return -1;
  if (src == NULL || *src == '\0')
    return 0;

  /* How long is src ?
   */
  src_len  = strlen(src);

  if (siz > 0) {

    /* How much free space do we have ?
     */
    dst_end  = strlen(dst);
    dst_free = siz - dst_end;

    /* Copy at most dst_free chars.
     * The documentation on strncat() is inconclusive as to
     * whether the NULL counts or not, thus use strncpy().
     * dst[dst_end] is the NULL terminator of dst.
     */
    (void) strncpy(&dst[dst_end],
                   src,
                   ( (src_len < dst_free) ? src_len : dst_free));

    /* NULL terminate dst.
     */
    if (src_len < dst_free)
      dst[dst_end+src_len] = '\0';
    dst[siz-1] = '\0';

    if (src_len >= dst_free)
      return 1;
  }

  return 0;
}

/*
 * An alternative implementation of the OpenBSD strlcpy() function.
 *
 * Copy src to string dst of size siz.  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 */
static int sl_strlcpy(char * dst, const char * src, size_t siz)
{
  if (dst == NULL)
    return 0;
  if (src == NULL)
    {
      if (siz > 0)
        dst[0] = '\0';
      return 0;
    }


  if (siz > 0) {
    /* copy siz-1 characters
     */
    (void) strncpy(dst, src, siz-1);

    /* NULL terminate
     */
    dst[siz-1] = '\0';
  }
  return 0;
}

static int sh_util_hexchar( char c )
{
  /*@+charint@*/
  if      ( c >= '0' && c <= '9' )
    return (int)(c - '0');
  else if ( c >= 'a' && c <= 'f' )
    return (int)(c - 'a' + 10);
  else if ( c >= 'A' && c <= 'F' )
    return (int)(c - 'A' + 10);
  else return (int)('?');
  /*@-charint@*/
}

static char i2h[2];

static char * int2hex (unsigned char i)
{
  int j, k;

  j = (int)i / 16;
  k = (int)i - (j*16);
  /*@+charintliteral+charint@*/
  if (j < 10)
    i2h[0] = '0'+j;
  else
    i2h[0] = 'A'+(j-10);

  if (k < 10)
    i2h[1] = '0'+k;
  else
    i2h[1] = 'A'+(k-10);
  /*@-charintliteral-charint@*/

  return i2h;
}

static size_t add_check (int line, size_t a, size_t b)
{
  if (a > (SIZE_MAX - b))
    die(line, "Integer overflow");
  return (a+b);
}

static char * quote_string (char * str)
{
  size_t i = 0, j, k = 0, len;
  char * tmp;
  char * tmp2;


  if (str == NULL)
    {
      die (__LINE__, "NULL input string\n");
    }

  len = strlen(str);

  for (j = 0; j < len; ++j)
    if (str[j] == '\n' || str[j] == QUOTE_CHAR) ++i;

  j = add_check(__LINE__, len, i);
  j = add_check(__LINE__, j,   i);
  j = add_check(__LINE__, j,   i);
  j = add_check(__LINE__, j,   1);

  /* tmp = xalloc (__LINE__, len + 1 + 3*i); */
  tmp = xalloc (__LINE__, j);

  for (j = 0; j <= len; ++j)
    {
      if (str[j] == '\n')
        {
          tmp2 = int2hex((unsigned char) '\n');
          tmp[k] = QUOTE_CHAR; ++k;
          tmp[k] = tmp2[0];    ++k;
          tmp[k] = tmp2[1];
        }
      else if (str[j] == QUOTE_CHAR)
        {
          tmp2 = int2hex((unsigned char) QUOTE_CHAR);
          tmp[k] = QUOTE_CHAR; ++k;
          tmp[k] = tmp2[0];    ++k;
          tmp[k] = tmp2[1];
        }
      else
        {
          tmp[k] = str[j];
        }
      ++k;
    }
  return tmp;
}

static char * unquote_string (char * str)
{
  int i = 0; 
  size_t j, k = 0, len;
  char * tmp;

  if (str == NULL)
    {
     return NULL;
    }

  len = strlen(str);

  j = add_check(__LINE__, len, 1);
  tmp = xalloc (__LINE__, j);

  for (j = 0; j <= len; ++j)
    {
      if ((str[j] == QUOTE_CHAR) && (len > 2) && (j < (len-2)))
        {
          if (sh_util_hexchar(str[j+1]) >= 0 && sh_util_hexchar(str[j+2]) >= 0)
            {
              i = 16 * sh_util_hexchar(str[j+1]) + sh_util_hexchar(str[j+2]);
              tmp[k] = (char) i; j += 2;
            }
          else
            {
              tmp[k] = str[j];
            }
        }
      else
        tmp[k] = str[j];
      ++k;
    }
  return tmp;
}

static int sh_unix_getline (FILE * fd, char * line, int sizeofline)
{
  register int  n = 0;
  char        * res;

  if (sizeofline < 2) {
    line[0] = '\0';
    return 0;
  }
  res = fgets(line, sizeofline, fd);
  if (res == NULL)
    {
      line[0] = '\0';
      return -1;
    }
  n = strlen(line);
  if (n > 0 && line[n-1] == '\n') {
    --n; line[n] = '\0';
  }
  return n;
}

static char * make_lockfile (const char * string)
{
  char * targetfile = NULL;
  size_t  i      = strlen(string);

  i = add_check(__LINE__, i, 5);
  targetfile = xalloc (__LINE__, i);

  (void) sl_strlcpy(targetfile, string,           i);
  (void) sl_strlcat(targetfile, ".lck",           i);
  return targetfile;
}

static int lockfile_fd = -1;

static void create_lock (const char * string)
{
  struct flock lock;
  char * lockfile = make_lockfile (string);

  lockfile_fd = open (lockfile, O_WRONLY|O_CREAT, 
		      S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
  if (lockfile_fd < 0)
    die(__LINE__, "Cannot open file <%s>: %s\n", 
	lockfile, strerror(errno));

  lock.l_type   = F_WRLCK;
  lock.l_whence = SEEK_SET;
  lock.l_start  = 0;
  lock.l_len    = 0;

  // F_SETLKW blocks until lock is released

  fcntl(lockfile_fd, F_SETLKW, &lock);
  return;
}

static void remove_lock ()
{
  int res = -1;

  if (lockfile_fd < 0)
    return;

  do {
    res = close(lockfile_fd);
  } while (res == -1 && errno == EINTR);

  lockfile_fd = -1;
  return;
}


/**************************************************************
 *
 * Hashtable functions
 *
 **************************************************************/

#define TABSIZE 65536
static sh_file_t * hashtable[TABSIZE];
static int hash_is_init = 0;

/*
--------------------------------------------------------------------
mix -- mix 3 32-bit values reversibly.
For every delta with one or two bits set, and the deltas of all three
  high bits or all three low bits, whether the original value of a,b,c
  is almost all zero or is uniformly distributed,
* If mix() is run forward or backward, at least 32 bits in a,b,c
  have at least 1/4 probability of changing.
* If mix() is run forward, every bit of c will change between 1/3 and
  2/3 of the time.  (Well, 22/100 and 78/100 for some 2-bit deltas.)
mix() was built out of 36 single-cycle latency instructions in a 
  structure that could supported 2x parallelism, like so:
      a -= b; 
      a -= c; x = (c>>13);
      b -= c; a ^= x;
      b -= a; x = (a<<8);
      c -= a; b ^= x;
      c -= b; x = (b>>13);
      ...
  Unfortunately, superscalar Pentiums and Sparcs can't take advantage 
  of that parallelism.  They've also turned some of those single-cycle
  latency instructions into multi-cycle latency instructions.  Still,
  this is the fastest good hash I could find.  There were about 2^^68
  to choose from.  I only looked at a billion or so.
--------------------------------------------------------------------
*/
#define mix(a,b,c)                              \
{                                               \
  a -= b; a -= c; a ^= (c>>13);                 \
  b -= c; b -= a; b ^= (a<<8);                  \
  c -= a; c -= b; c ^= (b>>13);                 \
  a -= b; a -= c; a ^= (c>>12);                 \
  b -= c; b -= a; b ^= (a<<16);                 \
  c -= a; c -= b; c ^= (b>>5);                  \
  a -= b; a -= c; a ^= (c>>3);                  \
  b -= c; b -= a; b ^= (a<<10);                 \
  c -= a; c -= b; c ^= (b>>15);                 \
}

/*
--------------------------------------------------------------------
hash() -- hash a variable-length key into a 32-bit value
  k       : the key (the unaligned variable-length array of bytes)
  len     : the length of the key, counting by bytes
  initval : can be any 4-byte value
Returns a 32-bit value.  Every bit of the key affects every bit of
the return value.  Every 1-bit and 2-bit delta achieves avalanche.
About 6*len+35 instructions.

The best hash table sizes are powers of 2.  There is no need to do
mod a prime (mod is sooo slow!).  If you need less than 32 bits,
use a bitmask.  For example, if you need only 10 bits, do
  h = (h & hashmask(10));
In which case, the hash table should have hashsize(10) elements.

If you are hashing n strings (ub1 **)k, do it like this:
  for (i=0, h=0; i<n; ++i) h = hash( k[i], len[i], h);

By Bob Jenkins, 1996.  bob_jenkins@burtleburtle.net.  You may use this
code any way you wish, private, educational, or commercial.  It's free.

See http://burtleburtle.net/bob/hash/evahash.html
Use for hash table lookup, or anything where one collision in 2^^32 is
acceptable.  Do NOT use for cryptographic purposes.
--------------------------------------------------------------------
*/

typedef UINT32 ub4; 

static UINT32
hash_func(void *key, size_t length)
{
    register ub4 a,b,c,len;
    char *k = key;

    /* Set up the internal state */
    len = length;
    a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
    c = 0;               /* the previous hash value */

    /*---------------------------------------- handle most of the key */
    while(len >= 12){
        a += (k[0] +((ub4)k[1]<<8) +((ub4)k[2]<<16) +((ub4)k[3]<<24));
        b += (k[4] +((ub4)k[5]<<8) +((ub4)k[6]<<16) +((ub4)k[7]<<24));
        c += (k[8] +((ub4)k[9]<<8) +((ub4)k[10]<<16)+((ub4)k[11]<<24));
        mix(a,b,c);
        k += 12; len -= 12;
    }

    /*------------------------------------- handle the last 11 bytes */
    c += length;
    switch(len){              /* all the case statements fall through */
    case 11: c+=((ub4)k[10]<<24);
    case 10: c+=((ub4)k[9]<<16);
    case 9 : c+=((ub4)k[8]<<8);
        /* the first byte of c is reserved for the length */
    case 8 : b+=((ub4)k[7]<<24);
    case 7 : b+=((ub4)k[6]<<16);
    case 6 : b+=((ub4)k[5]<<8);
    case 5 : b+=k[4];
    case 4 : a+=((ub4)k[3]<<24);
    case 3 : a+=((ub4)k[2]<<16);
    case 2 : a+=((ub4)k[1]<<8);
    case 1 : a+=k[0];
        /* case 0: nothing left to add */
    }
    mix(a,b,c);
    /*-------------------------------------------- report the result */
    return (c % TABSIZE);
}

/* End of code from Jenkins */


static void hash_init ()
{
  int i;
  for (i = 0; i < TABSIZE; ++i)
    hashtable[i] = NULL;
  hash_is_init = 1;
  return;
}

void hash_free (sh_file_t * p) 
{
  if (p->fullpath)
    free (p->fullpath);
  p->fullpath = NULL;
  if (p->linkpath)
    free (p->linkpath);
  p->linkpath = NULL;
  if (p->attr_string)
    free (p->attr_string);
  p->attr_string = NULL;
  free(p);
}


static void hash_kill (sh_file_t *p)
{
  if (p == NULL)
    return;
  if (p->next != NULL)
    hash_kill (p->next);
  hash_free(p);
  return;
}

static void hash_destroy ()
{
  int i;

  if (hash_is_init == 0) 
    return;

  for (i = 0; i < TABSIZE; ++i) 
    if (hashtable[i] != NULL)
      { 
        hash_kill (hashtable[i]);
        hashtable[i] = NULL;
      }
  hash_is_init = 0;
  return;
}

static void hash_replace(sh_file_t * old, sh_file_t * new, rFlag read_flag)
{
  sh_file_t * q;

  q = old->next;

  free(old->fullpath); 
  if(old->linkpath)    
    free(old->linkpath);
  if(old->attr_string) 
    free(old->attr_string); 

  if (read_flag != read_merge)
    {
      memcpy(old, new, sizeof(sh_file_t));
    }
  else
    {
      UINT32 checkflags = old->theFile.checkflags;
      memcpy(old, new, sizeof(sh_file_t));
      if (checkflags != 0)
	old->theFile.checkflags = checkflags;
    }
  old->fullpath = xstrdup(__LINE__, new->fullpath);
  if (new->linkpath)    { old->linkpath    = xstrdup(__LINE__, new->linkpath); }
  else old->linkpath = NULL;
  if (new->attr_string) { old->attr_string = xstrdup(__LINE__, new->attr_string); }
  else old->attr_string = NULL;
  old->next = q;
  return;
}

static int hash_probe (sh_file_t * p, rFlag read_flag)
{
  sh_file_t * test;

  if (hash_is_init == 0)
    hash_init();

  if (hashtable[p->hashkey] == NULL)
    {
      hashtable[p->hashkey] = p;
      return 1;
    }
  else
    {
      test = hashtable[p->hashkey];

      while (test)
	{
	  if (strcmp(test->fullpath, p->fullpath) == 0)
	    {
	      hash_replace(test, p, read_flag);
	      return 0;
	    }
	  if (test->next == NULL)
	    break;
	  test = test->next;
	}

      test->next = p;
      return 1;
    }
}

static int hash_delete (char * path)
{
  sh_file_t * test;
  sh_file_t * prev;
  size_t len = strlen(path);
  UINT32 key = hash_func (path, len);

  test = hashtable[key];
  prev = test;

  while (test)
    {
      if (0 == strcmp(test->fullpath, path))
	{
	  if (test == hashtable[key])
	    hashtable[key] = test->next;
	  else
	    prev->next     = test->next;
	  hash_free(test);
	  return 1;
	}
      prev = test;
      test = test->next;
    }
  return 0;
}

static sh_file_t * hash_iter_curr  = NULL;
static long        hash_iter_count = 0; 

static void hash_iter_start()
{
  hash_iter_count = 0;
  hash_iter_curr  = NULL;
  return;
}

/* each call to this function will return the next entry
*/
static sh_file_t * hash_iter()
{
  if ((hash_iter_curr != NULL) && (hash_iter_curr->next != NULL))
    {
      hash_iter_curr = hash_iter_curr->next;
      return hash_iter_curr;
    }
  else
    {
      if (hash_iter_count >= TABSIZE)
	return NULL;

      while (hashtable[hash_iter_count] == NULL)
	{
	  ++hash_iter_count;
	  if (hash_iter_count >= TABSIZE)
	    return NULL;
	}

      hash_iter_curr = hashtable[hash_iter_count];
      ++hash_iter_count;

      return hash_iter_curr;
    }
}

int sh_hash_is_null_record(sh_filestore_t * theFile)
{
  if (theFile->hardlinks == SH_DEADFILE && theFile->mode  == 0 &&
      theFile->ino == 0                 && theFile->ctime == 0)
    {
      return 1;
    }
  return 0;
}


/******************************************************************
 *
 * Fast forward to start of data
 *
 ******************************************************************/
static int sh_hash_setdataent (FILE * fd, char * line, char * database_name,
			       FILE * newfile)
{
  long i;
  int  pgp_flag = 0;
  int  werr;

  rewind (fd);

  while (1)
    {
      i =  sh_unix_getline (fd, line, SH_BUFSIZE-1);
      if (i < 0 )
	die (__LINE__, "No data in file %s\n", database_name);

#if defined(SH_STEALTH)
      if (0 == strncmp (line, N_("[SOF]"), 5))
#else
      if (0 == strncmp (line, _("[SOF]"),  5))
#endif
	{
	  pverb(1, __LINE__, _("Found SOF marker (%s)\n"), _(line));
	  break;
	}

      /* -- write everything until SOF marker to new file
       * -- don't write PGP marker lines
       */
      if (newfile != NULL)
	{
	  if (line[0] == '\0')
	    {
	      if (pgp_flag == 0)
		fputc('\n', newfile);
	      pgp_flag = 0;
	    }
	  else if (i == (SH_BUFSIZE-1))
	    {
	      if (0 != strncmp(line, "-----", 5))
		{
		  werr = fwrite(line, i, 1, newfile);
		  CHECK_ERR(werr);
		}
	      pgp_flag = 0;
	    }
	  else
	    {
	      if (i < 5)
		{
		  if (i > 0)
		    {
		      werr = fwrite(line, i, 1, newfile);
		      CHECK_ERR(werr);
		    }
		  fputc('\n', newfile);
		  pgp_flag = 0;
		}
	      else if ( (0 == strncmp(line, "-----", 5)) ||
			(0 == strncmp(line, "Hash:", 5)) ||
			(0 == strncmp(line, "NotDa", 5)))
		{
		  pgp_flag = 1;
		}
	      else
		{
		  werr = fwrite(line, i, 1, newfile);
		  CHECK_ERR(werr);
		  fputc('\n', newfile);
		  pgp_flag = 0;
		}
	    }
	}
    }
  return 1;
}

static size_t dbIO_fread_struct (sh_filestore_t * ptr, FILE *stream, 
				 const char * path, int * isNew)
{
  sh_filestore_old_t old_struct;
  fpos_t position;
  static int oldflag = -1;

 start:
  if (oldflag != -1) /* 'initialized' case first */
    {
      if (oldflag == 0)
	return fread (ptr, sizeof(sh_filestore_t), 1, stream);
      else
	{
	  unsigned short mark;
	  if (1 != fread (&old_struct, sizeof(old_struct), 1, stream))
	    return 0;
	  /* set mark to current version */
	  mark = old_struct.mark;
	  mark = *(swap_short(&(mark)));
	  if ((mark & ~REC_FLAGS_MASK) != OLD_REC_MAGIC)
	    {
	      return 0;
	    }
	  if ((mark & REC_FLAGS_ATTR) != 0)
	    mark = REC_MAGIC|REC_FLAGS_ATTR;
	  else
	    mark = REC_MAGIC;
	  mark = *(swap_short(&(mark)));
	  old_struct.mark = mark;
	  /* copy into current struct version */
	  memcpy(ptr, &old_struct, sizeof(old_struct));
	  ptr->checkflags = 0;
	  return 1;
	}
    }
  else /* not initialized yet, test DB version */
    {
      if (0 == fgetpos(stream, &position))
	{
	  unsigned short mark;
	  if (1 != fread (&old_struct, sizeof(old_struct), 1, stream))
	    return 0;
	  mark = old_struct.mark;
	  mark = *(swap_short(&(mark)));
	  if ((mark & ~REC_FLAGS_MASK) == REC_MAGIC)
	    { oldflag = 0; *isNew = 1; }
	  else if ((mark & ~REC_FLAGS_MASK) == OLD_REC_MAGIC)
	    { oldflag = 1; *isNew = 0; }
	  else
	    die (__LINE__, _("Invalid record in file %s\n"), path);
	  /* return to previous position and read data */
	  if (0 != fsetpos(stream, &position))
	    return 0;
	  goto start;
	}
      return 0;
    }
}


/******************************************************************
 *
 * Read next record
 *
 ******************************************************************/
static sh_file_t *  sh_hash_getdataent (FILE * fd, char * line, 
					char * database_name, int * isNew)
{
  sh_file_t * p;
  sh_filestore_t ft;
  long i;
  char * fullpath    = NULL;
  char * linkpath    = NULL;
  char * attr_string = NULL;
  char * tmp         = NULL;
  size_t pathlen;

  /* Read next record -- Part One
   */
  p = xalloc (__LINE__, sizeof(sh_file_t));

  i = dbIO_fread_struct(&ft, fd, database_name, isNew);
  if ( i < 1)
    {
      pverb(1, __LINE__, _("End of database reached\n"));
      free (p);
      return NULL;
    }
  p->fullpath = NULL;
  p->linkpath = NULL;

  swap_32(&(ft.mode));
  swap_32(&(ft.linkmode));
  swap_64(&(ft.dev));
  swap_64(&(ft.rdev));
  swap_32(&(ft.hardlinks));
  swap_32(&(ft.ino));
  swap_64(&(ft.size));
  swap_64(&(ft.atime));
  swap_64(&(ft.mtime));
  swap_64(&(ft.ctime));
  swap_32(&(ft.owner));
  swap_32(&(ft.group));
  swap_32(&(ft.checkflags));
#if defined(__linux__)
  swap_32(&(ft.attributes));
#endif

  ft.mark = *(swap_short(&(ft.mark)));

  if ((ft.mark & ~REC_FLAGS_MASK) != REC_MAGIC)
    {
      pverb(1, __LINE__, _("End of valid database records\n"));
      free (p);
      return NULL;
    }

  /* Read next record -- Part Two -- Fullpath
   */
  i =  sh_unix_getline (fd, line, SH_BUFSIZE-1);
  if (i < 0 )
    die (__LINE__, _("No data in file %s\n"), database_name);

  tmp = unquote_string (line);
  if (tmp)
    {
      i = strlen(tmp)+1;
      fullpath = xalloc(__LINE__, i);
      sl_strlcpy (fullpath, tmp, i);
      free (tmp);
    }
  else
    {
      i = 2;
      fullpath = xalloc(__LINE__, i);
      sl_strlcpy (fullpath, "-", i);
    }

  if (fullpath[i-2] == '\n')
    {
      fullpath[i-2] = '\0';
      --i;
    }

  pathlen = i-1;

  /* Read next record -- Part Three -- Linkpath
   */
  i =  sh_unix_getline (fd, line, SH_BUFSIZE-1);
  if (i < 0 )
    die (__LINE__, "No data in file %s\n", database_name);

  tmp = unquote_string (line);
  if (tmp)
    {
      i = strlen(tmp)+1;
      linkpath = xalloc(__LINE__, i);
      sl_strlcpy (linkpath, tmp, i);
      free (tmp);
    }
  else
    {
      i = 2;
      linkpath = xalloc(__LINE__, i);
      sl_strlcpy (linkpath, "-", i);
    }
  if (linkpath[i-2] == '\n')
    linkpath[i-2] = '\0';

  /* Read next record -- Part Four -- attr_string
   */
  if ((ft.mark & REC_FLAGS_ATTR) != 0)
    {
      i =  sh_unix_getline (fd, line, SH_BUFSIZE-1);
      if (i < 0 )
	die (__LINE__, "No data in file %s\n", database_name);
      
      tmp = unquote_string (line);
      if (tmp)
	{
	  i = strlen(tmp)+1;
	  attr_string = xalloc(__LINE__, i);
	  sl_strlcpy (attr_string, tmp, i);
	  free (tmp);
	  if (attr_string[i-2] == '\n')
	    attr_string[i-2] = '\0';
	}
    }
 
  /* Read next record -- Part Five -- Decode
   */
#if defined(SH_STEALTH)
  sh_do_decode(fullpath,    strlen(fullpath));

#if defined(__linux__)
  sh_do_decode(ft.c_attributes,   strlen(ft.c_attributes));
#endif

  sh_do_decode(ft.c_mode,   strlen(ft.c_mode));
  sh_do_decode(ft.c_owner,  strlen(ft.c_owner));
  sh_do_decode(ft.c_group,  strlen(ft.c_group));
  sh_do_decode(ft.checksum, strlen(ft.checksum));


  if (ft.c_mode[0] == 'l' && linkpath[0] != '-')
    {
      sh_do_decode(linkpath, strlen(linkpath));
    }
  if (attr_string)
    {  
      sh_do_decode(attr_string, strlen(attr_string));
    }
#endif

  memcpy( &(*p).theFile, &ft, sizeof(sh_filestore_t) );
  p->visited   = S_FALSE;
  p->reported  = S_FALSE;
  p->allignore = S_FALSE;
  p->modi_mask = 0L;
  p->fullpath  = fullpath;
  p->linkpath  = linkpath;

  p->attr_string  = attr_string;

  /* set to an invalid value
   */
  ft.mark = (REC_MAGIC + 5);

  p->hashkey = hash_func(fullpath, pathlen);
  p->next    = 0;

  return p;
}

/************************************************************
 *
 * Read database in memory and operate on it.
 *
 ************************************************************/

/* nr. of records in database
 */
static int         DATA_RECD   = 0;

/* name of database file
 */
static char      * DATA_CURR = NULL;

static void sh_hash_pushdata (FILE * fd, sh_file_t * p);
static FILE * safe_open (char * to_file, char * mode, int world_readable);
static int make_backup (char * target);

#ifdef NO_DB_BACKUP
static int remove_backup (char * target);
#endif

/*@-exportlocal@*/ void write_database(char * db_file);
/*@-exportlocal@*/ void close_database(char * db_file);


static int read_database (char * db_file, rFlag read_flag)
{ 
  int i = 0;
  sh_file_t * p;
  FILE * fd;
  char * line;
  char * newfile = NULL;

  if (!db_file)
    die(__LINE__, _("Database is NULL\n")); 

  if (DATA_CURR == NULL)
    {
      DATA_CURR = xalloc (__LINE__, strlen(db_file) + 1);
      strcpy(DATA_CURR, db_file);
    }

  line = xalloc (__LINE__, SH_BUFSIZE+1);

  /* Place a lock on the lockfile. The lock is removed
   * in close_database().
   */
  if (read_flag == read_replace)
    create_lock (db_file);

  if ( NULL == (fd = fopen(db_file, "r"))) 
    {
      /* perror("fopen"); */
      die(__LINE__, _("Can't open %s for read: %s\n"), 
	  db_file, strerror(errno));
    }

  /*
   * Keep f_tmp if we want to merge
   */
  if (read_flag == read_replace)
    {
      newfile = xalloc (__LINE__,    1 + strlen(db_file) + 4);
      sl_strlcpy(newfile, db_file,   1 + strlen(db_file) + 4);
      sl_strlcat(newfile, _(".new"), 1 + strlen(db_file) + 4);
  
      f_tmp = safe_open(newfile, "w", 0);
      if (!f_tmp)
	die(__LINE__, _("Can't open %s for write\n"), newfile);
    }

  /* Fast forward to start of data, and do NOT write to f_tmp if
   * this is the DB to be merged
   */
  if  (read_flag == read_replace)
    sh_hash_setdataent(fd, line, db_file, f_tmp);
  else
    sh_hash_setdataent(fd, line, db_file, NULL);

  /* Write start marker to f_tmp unless this is the
   * DB to be merged
   */
  if (read_flag == read_replace)
    {
      fprintf (f_tmp, N_("[SOF]"));
      fprintf (f_tmp, "%s", _("\n"));
      (void) fflush  (f_tmp);
    }

  while (1) 
    {
      if (read_flag == read_merge)
	p = sh_hash_getdataent (fd, line, db_file, &ISNEW_MERGE);
      else
	p = sh_hash_getdataent (fd, line, db_file, &ISNEW_CURR);
      if (p != NULL)
	{
	  if (read_flag == read_merge)
	    pverb(2, __LINE__, _("merging %s\n"), p->fullpath);
	  if (!sh_hash_is_null_record(&(p->theFile)))
	    {
	      if (hash_probe (p, read_flag) > 0)
		{
		  if (read_flag == read_merge)
		    pverb(2, __LINE__, _("--> new\n"));
		  ++i;
		}
	      else /* path was present */
		{
		  if (read_flag == read_merge)
		    pverb(2, __LINE__, _("--> replace old\n"));
		  hash_free(p);
		}
	    }
	  else
	    {
	      if (0 != hash_delete(p->fullpath))
		{
		  if (read_flag == read_merge)
		    pverb(2, __LINE__, _("--> removed\n"));
		  --DATA_RECD;
		}
	      else
		{
		  if (read_flag == read_merge)
		    pverb(2, __LINE__, _("--> to remove but not present\n"));
		}
	    }
	} 
      else
	{
	  break;
	}
    }

  if (line != NULL)
    free(line);
  fclose (fd);

  if (read_flag == read_replace)
    DATA_RECD = i;
  else
    DATA_RECD += i;

  if (newfile)
    free(newfile);
  return 0;
}


void write_database(char * db_file)
{
  int i = 0;
  sh_file_t * p;

  if (!f_tmp)
    return;

  hash_iter_start();

  while (NULL != (p = hash_iter()))
    {
      sh_hash_pushdata (f_tmp, p);
      ++i;
    }
  hash_destroy();

  if (i != DATA_RECD)
    {
      die (__LINE__, 
	   "database %s: %d expected, but %d records found\n",
	   db_file, DATA_RECD, i);
    }

#ifdef SYSLOG_DBG
  syslog (SYSLOG_MODE, "write_database: DATA_RECD %d", (int) DATA_RECD);
#endif

  DATA_RECD = 0;
  return;
}

void close_database(char * db_file)
{
  char * newfile;
#ifdef HAVE_FILENO
  int    fd;
#endif

  if (!f_tmp)
    return;

  if (!db_file)
    die(__LINE__, _("Database is NULL\n")); 

  newfile = xalloc (__LINE__,    1 + strlen(db_file) + 4);
  sl_strlcpy(newfile, db_file,   1 + strlen(db_file) + 4);
  sl_strlcat(newfile, _(".new"), 1 + strlen(db_file) + 4);
  (void) fflush (f_tmp);
#ifdef HAVE_FILENO
  fd = fileno(f_tmp);
  if (fd >= 0) {
#ifdef HAVE_FDATASYNC
    fdatasync(fd);
#else
    fsync(fd);
#endif
  }
#endif
  fclose (f_tmp);
  f_tmp = NULL;

  /* create backup
   */
  make_backup(db_file);

  /* rename temporary file to database
   */
  if (0 != rename (newfile, db_file))
    die (__LINE__, _("Could not rename %s to %s: %s\n"), 
	 newfile, db_file, strerror(errno));

#ifdef NO_DB_BACKUP
  /* remove the backup
   */
  remove_backup(db_file);
#endif

  /* Remove the lock on the lockfile.
   */
  remove_lock();

  free (newfile);
  return;
}

/******************************************************************
 *
 * Merge database
 *
 ******************************************************************/
static void merge_database(char * db_file, char * db_merge)
{
  pverb(1, __LINE__, _("Merge %s into %s\n"), db_merge, db_file);
  read_database(db_file, read_replace);
  read_database(db_merge, read_merge);

  if (ISNEW_CURR != ISNEW_MERGE)
    die(__LINE__, _("Attempt to merge databases of different versions!\n"));

  pverb(1, __LINE__, _("Write out %s\n"), DATA_CURR);
  write_database(DATA_CURR);
  close_database(DATA_CURR);
  return;
}


/******************************************************************
 *
 * Update database
 *
 ******************************************************************/
static void sh_hash_pushdata (FILE * fd, sh_file_t * p)
{
  char * tmp;
  static char fullpath[MAX_PATH_STORE+1];
  static char linkpath[MAX_PATH_STORE+1];
  char *  attr_string = NULL;
#if defined(SH_STEALTH)
  size_t old_len;
#endif
  size_t tmp_len;
  int    werr;

  linkpath[0] =  '-';
  linkpath[1] = '\0';
  fullpath[0] =  '-';
  fullpath[1] = '\0';

#if defined(SH_STEALTH)
  old_len = strlen(p->fullpath);
  sh_do_encode(p->fullpath, old_len);
#endif
  tmp = quote_string(p->fullpath);
  tmp_len = strlen(tmp);
#if defined(SH_STEALTH)
    sh_do_decode(p->fullpath, old_len);
#endif
  if (tmp_len <= MAX_PATH_STORE)
    {
      sl_strlcpy(fullpath, tmp, MAX_PATH_STORE+1);
    }
  free(tmp);


  if (p->theFile.c_mode[0] == 'l' || 
      (p->linkpath != NULL && p->linkpath[0] != '-'))
    {
#if defined(SH_STEALTH)
      old_len = strlen(p->linkpath);
      if (p->theFile.c_mode[0] == 'l')
	sh_do_encode(p->linkpath, old_len);
#endif
      tmp = quote_string(p->linkpath);
      tmp_len = strlen(tmp);
#if defined(SH_STEALTH)
      if (p->theFile.c_mode[0] == 'l')
	sh_do_decode(p->linkpath, old_len);
#endif

      if (tmp_len <= MAX_PATH_STORE)
	{
	  sl_strlcpy(linkpath, tmp, MAX_PATH_STORE+1);
	}
      free(tmp);
    }

  if (p->attr_string != NULL)
    {
#if defined(SH_STEALTH)
      old_len = strlen(p->attr_string);
      sh_do_encode(p->attr_string, old_len);
#endif
      tmp = quote_string(p->attr_string);
      tmp_len = strlen(tmp);
#if defined(SH_STEALTH)
      sh_do_decode(p->attr_string, old_len);
#endif

      attr_string = xalloc (__LINE__, tmp_len + 1);
      sl_strlcpy(attr_string, tmp, tmp_len + 1);
      free(tmp);
    }

#if defined(NO_DB_UPGRADE)
  if (ISNEW_CURR == 1)
    p->theFile.mark = REC_MAGIC;
  else
    p->theFile.mark = OLD_REC_MAGIC;
#else
  p->theFile.mark = REC_MAGIC;
#endif

  if (attr_string)
    p->theFile.mark |= REC_FLAGS_ATTR;

#if defined(SH_STEALTH)
  sh_do_encode(p->theFile.c_mode,   strlen(p->theFile.c_mode));
  sh_do_encode(p->theFile.c_owner,  strlen(p->theFile.c_owner));
  sh_do_encode(p->theFile.c_group,  strlen(p->theFile.c_group));
  sh_do_encode(p->theFile.checksum, strlen(p->theFile.checksum));
#if defined(__linux__)
  sh_do_encode(p->theFile.c_attributes,   strlen(p->theFile.c_attributes));
#endif
#endif

  swap_32(&(p->theFile.mode));
  swap_32(&(p->theFile.linkmode));
  swap_64(&(p->theFile.dev));
  swap_64(&(p->theFile.rdev));
  swap_32(&(p->theFile.hardlinks));
  swap_32(&(p->theFile.ino));
  swap_64(&(p->theFile.size));
  swap_64(&(p->theFile.atime));
  swap_64(&(p->theFile.mtime));
  swap_64(&(p->theFile.ctime));
  swap_32(&(p->theFile.owner));
  swap_32(&(p->theFile.group));
  swap_32(&(p->theFile.checkflags));
#if defined(__linux__)
  swap_32(&(p->theFile.attributes));
#endif
  p->theFile.mark = *(swap_short(&(p->theFile.mark)));

#if defined(NO_DB_UPGRADE)
  if (ISNEW_CURR == 1)
    werr = fwrite      (&(p->theFile), sizeof(sh_filestore_t), 1, fd);
  else
    werr = fwrite      (&(p->theFile), sizeof(sh_filestore_old_t), 1, fd);
#else
  werr = fwrite      (&(p->theFile), sizeof(sh_filestore_t), 1, fd);
#endif
  CHECK_ERR(werr);
  werr = fwrite      (fullpath, strlen(fullpath), 1, fd);
  CHECK_ERR(werr);
  werr = fwrite      ("\n", 1, 1, fd);
  CHECK_ERR(werr);
  werr = fwrite      (linkpath, strlen(linkpath), 1, fd);
  CHECK_ERR(werr);
  werr = fwrite      ("\n", 1, 1, fd);
  CHECK_ERR(werr);
  if (attr_string)
    {
      werr = fwrite      (attr_string, strlen(attr_string), 1, fd);
      CHECK_ERR(werr);
      werr = fwrite      ("\n", 1, 1, fd);
      CHECK_ERR(werr);
      free (attr_string);
    }
 
  return;
}

static int is_group_writeable = 0;

static FILE * safe_open (char * to_file, char * mode, int world_readable)
{
  struct stat lbuf;
  struct stat fbuf;
  int fd_t = -1;
  FILE * fp;
  mode_t mask;

  if (0 != lstat(to_file, &lbuf))
    {
      if (errno == ENOENT)
	{
	  if (0 != world_readable)
	    {
	      mask = umask(0);
	      if (is_group_writeable == 0)
		{
		  fd_t = open (to_file, O_WRONLY|O_CREAT|O_EXCL, 
			       S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
		}
	      else
		{
		  fd_t = open (to_file, O_WRONLY|O_CREAT|O_EXCL, 
			       S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH);
		}
	      umask(mask);
	    }
	  else
	    {
	      fd_t = open (to_file, O_WRONLY|O_CREAT|O_EXCL, 
			   S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH);
	    }
	  if (fd_t < 0)
	    die(__LINE__, _("Cannot open file <%s>: %s\n"), 
		to_file, strerror(errno));
	} else {
	  die(__LINE__, _("Cannot lstat file <%s>: %s\n"), 
		to_file, strerror(errno));
	}
    }
  else
    {
      if (!S_ISREG(lbuf.st_mode) && entry_undo == 0)
	die(__LINE__, _("Not a regular file <%s>\n"), to_file);
      if (mode[0] == 'a')
	fd_t = open (to_file, O_WRONLY|O_APPEND);
      else
	fd_t = open (to_file, O_WRONLY);
      if (fd_t < 0)
	die(__LINE__, _("Cannot open file <%s>: %s\n"), 
	    to_file, strerror(errno));
      if (0 != fstat(fd_t, &fbuf))
	die(__LINE__, _("Cannot stat file descriptor <%d>: %s\n"), 
	    fd_t, strerror(errno));

      if (fbuf.st_dev != lbuf.st_dev   || fbuf.st_ino  != lbuf.st_ino ||
	  fbuf.st_mode != lbuf.st_mode || fbuf.st_size != lbuf.st_size)
	die(__LINE__, _("File <%s> has changed\n"), to_file);

      if (mode[0] == 'w' && entry_undo == 0)
	{
	  if (0 != ftruncate(fd_t, 0))
	    die(__LINE__, _("Cannot ftruncate file <%s>: %s\n"), 
		to_file, strerror(errno));
	}
    }

  fp = fdopen(fd_t, mode);
  if (!fp) {
    die(__LINE__, _("Cannot fdopen file <%s>: %s\n"), 
	to_file, strerror(errno));
  }
  return fp;
}

static int do_copy (char * from_file, char * to_file)
{
  struct stat lbuf;
  struct stat fbuf;
  int fd_f;
  int fd_t = -1;
  char * readbuf[BUFSIZ];
  size_t count;
  size_t status;


  /* Open the file we copy FROM
   */
  if (0 != lstat(from_file, &lbuf))
    die(__LINE__, _("Cannot stat file <%s>: %s\n"), 
	from_file, strerror(errno));
  if (!S_ISREG(lbuf.st_mode))
    die(__LINE__, _("Not a regular file <%s>\n"), from_file);
  fd_f = open (from_file, O_RDONLY);
  if (fd_f < 0)
    die(__LINE__, _("Cannot open file <%s>: %s\n"), 
	from_file, strerror(errno));
  if (0 != fstat(fd_f, &fbuf))
    die(__LINE__, _("Cannot stat file descriptor <%d>: %s\n"), 
	fd_f, strerror(errno));

  if (fbuf.st_dev != lbuf.st_dev   || fbuf.st_ino  != lbuf.st_ino ||
      fbuf.st_mode != lbuf.st_mode || fbuf.st_size != lbuf.st_size)
    die(__LINE__, _("File <%s> has changed\n"), from_file);
    
  /* Open the file we copy TO
   */
  if (0 != lstat(to_file, &lbuf))
    {
      if (errno == ENOENT)
	{
	  fd_t = open (to_file, O_WRONLY|O_CREAT|O_EXCL, 
		       S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
	  if (fd_t < 0)
	    die(__LINE__, _("Cannot open file <%s>: %s\n"), 
		to_file, strerror(errno));
	} else {
	  die(__LINE__, _("Cannot lstat file <%s>: %s\n"), 
		to_file, strerror(errno));
	}
    }
  else
    {
      if (!S_ISREG(lbuf.st_mode))
	die(__LINE__, _("Not a regular file <%s>\n"), to_file);
      fd_t = open (to_file, O_WRONLY);
      if (fd_t < 0)
	die(__LINE__, _("Cannot open file <%s>: %s\n"), 
	    to_file, strerror(errno));
      if (0 != fstat(fd_t, &fbuf))
	die(__LINE__, _("Cannot stat file descriptor <%d>: %s\n"), 
	    fd_t, strerror(errno));

      if (fbuf.st_dev != lbuf.st_dev   || fbuf.st_ino  != lbuf.st_ino ||
	  fbuf.st_mode != lbuf.st_mode || fbuf.st_size != lbuf.st_size)
	die(__LINE__, _("File <%s> has changed\n"), to_file);
      if (0 != ftruncate(fd_t, 0))
	die(__LINE__, _("Cannot ftruncate file <%s>: %s\n"), 
	    to_file, strerror(errno));
    }

  do {
    count = read(fd_f, readbuf, BUFSIZ);
    if      (count == (size_t)-1 && errno == EINTR)
      continue;
    else if (count == (size_t)-1)
      die(__LINE__, _("Error reading from <%s>: %s\n"), 
	  from_file, strerror(errno));
    else if (count == 0)
      break;
    else
      {
	do {
	  status = write(fd_t, readbuf, count);
	  if (status == (size_t)-1 && errno != EINTR)
	    die(__LINE__, _("Error writing to <%s>: %s\n"), 
		to_file, strerror(errno));
	} while (status == (size_t)-1 && errno == EINTR);
      }
  } while (1 == 1);

#ifdef HAVE_FDATASYNC
  fdatasync(fd_t);
#else
  fsync(fd_t);
#endif

  if (0 != close(fd_t))
    die(__LINE__,
	_("Error closing <%s>: %s\n"), to_file, strerror(errno));
  close(fd_f);
  return 0;
}

#ifdef NO_DB_BACKUP
static int remove_backup (char * target)
{
  struct stat buf;
  char * oldfile;

  oldfile = xalloc (__LINE__,    1 + strlen(target) + 4);
  sl_strlcpy(oldfile, target,    1 + strlen(target) + 4);
  sl_strlcat(oldfile, _(".old"), 1 + strlen(target) + 4);

  if (0 == lstat(oldfile, &buf))
    {
      (void) unlink(oldfile);
    }
  free(oldfile);

  return 0;
}
#endif

static int make_backup (char * target)
{
  struct stat buf;
  char * oldfile;

  oldfile = xalloc (__LINE__,    1 + strlen(target) + 4);
  sl_strlcpy(oldfile, target,    1 + strlen(target) + 4);
  sl_strlcat(oldfile, _(".old"), 1 + strlen(target) + 4);
  
  if (0 != lstat(oldfile, &buf))
    {
      if (errno == ENOENT)
	{
	  do_copy (target, oldfile);
	  chmod (oldfile, S_IRUSR|S_IWUSR);
	  free(oldfile);
	  return 0;
	}
      else
	die(__LINE__, _("Could not stat backup file <%s>: %s\n"), 
	    oldfile, strerror(errno));
    }

  sl_strlcpy(oldfile, target,    1 + strlen(target) + 4);
  sl_strlcat(oldfile, _(".bak"), 1 + strlen(target) + 4);

  if (0 != lstat(oldfile, &buf))
    {
      if ( (time(NULL) - buf.st_mtime) > 86400) 
	{
	  do_copy (target, oldfile);
	  chmod (oldfile, S_IRUSR|S_IWUSR);
	  free(oldfile);
	  return 0;
	}
      else
	{
	  free(oldfile);
	  return 0;
	}
    }

  do_copy (target, oldfile);
  chmod (oldfile, S_IRUSR|S_IWUSR);
  free(oldfile);
  return 0;
}

#define UUID_SIZE 36

int sh_uuid_check(const char * in)
{
  int 		i;
  const char	*cp;

  if (strlen(in) != UUID_SIZE)
    return -1;
  for (i=0, cp = in; i <= UUID_SIZE; i++,cp++) {
    if ((i == 8) || (i == 13) || (i == 18) ||
	(i == 23)) {
      if (*cp == '-')
	continue;
      else
	return -1;
    }
    if (i== UUID_SIZE)
      if (*cp == 0)
	continue;
    if (!isxdigit(*cp))
      return -1;
  }
  return 0;
}

static mode_t file_test (char * filetype, char * file)
{
  struct stat buf;

  if (0 != lstat(file, &buf))
    die(__LINE__, _("Cannot stat() %s file <%s>: %s\n"), 
	filetype, file, strerror(errno));

  if (!S_ISREG(buf.st_mode))
    die(__LINE__, _("%s file <%s> is not a regular file\n"),
	filetype, file);

  if ( (buf.st_mode & S_IWOTH) != 0 )
    die(__LINE__, _("%s file <%s> is world writeable\n"),
	filetype, file);

  if ( (buf.st_mode & S_IWGRP) != 0 && (buf.st_gid != getegid()))
    die(__LINE__, 
	_("%s file <%s> is group writeable for another group\n"),
	filetype, file);

  return buf.st_mode;
}

static void build_merge_path()
{
  char * out;
  size_t size;

  if (!g_merge)
    return;

  if (sh_uuid_check(g_merge) < 0)
    die(__LINE__, _("Merge uuid refernce <%s> not a valid uuid\n"), g_merge);

  size = strlen(g_database) + strlen(g_merge) + 2;
  out = xalloc(__LINE__, size);
  sl_strlcpy (out, g_database, size);
  sl_strlcat (out, ".", size);
  sl_strlcat (out, g_merge, size);
  free(g_merge);
  g_merge = out;

  file_test(_("merge"), g_merge);
  return;
}

/******************************************************************
 *
 * --- M A I N ---
 *
 ******************************************************************/
int main (int argc, char ** argv)
{
  char * str;
  struct stat buf;
  char * p;

  program_name = argv[0];
  sh_getopt_get (argc, argv);

  if (NULL == getpwuid(getuid()))
    die(__LINE__, _("UID <%ld> not a valid user\n"), (long) getuid()); 

  if ((getuid() != 0) && (getuid() != DEFAULT_UID))
    die(__LINE__, _("UID <%ld> not allowed to run this program (only root and %ld are allowed)\n"), 
	(long) getuid(), (long) DEFAULT_UID);

  if (g_database == NULL)
    die(__LINE__, _("Database name not specified\n"));

  if (NULL != strchr(g_database, '/') &&
      0 != strncmp(g_database, _("tmp/"), 4))
    die(__LINE__, 
	_("Only subdirectory hierarchy <tmp/> allowed for database name\n"));

  p = strrchr(g_database, '/');
  if (!p) 
    p = g_database;
  else
    ++p;

  if (p)
    {
      if (!(*p)) die(__LINE__, _("Database name invalid\n"));
      if (0 != strncmp(p, _("file."), 5))
	die(__LINE__, _("Database file name must start with <file.>\n"));
      p = strrchr(g_database, '.');
      if (0 == strcmp(p, _(".bak")))
	die(__LINE__, _("Database file name must not end with <.bak>\n"));
      if (0 == strcmp(p, _(".log")))
	die(__LINE__, _("Database file name must not end with <.log>\n"));
    }

  if (NULL != strstr(g_database, _("../")))
    die(__LINE__, _("Database name contains <../>\n"));

  if (g_database[0] == '/')
    die(__LINE__, _("Database name contains leading </>\n"));

  if (0 != lstat(DEFAULT_DATAROOT, &buf))
    die(__LINE__, _("Cannot stat() data directory <%s>: %s\n"), 
	DEFAULT_DATAROOT, strerror(errno));

  if (!S_ISDIR(buf.st_mode))
    die(__LINE__, _("Data directory <%s> is not a directory\n"),
	DEFAULT_DATAROOT);

  if ( (buf.st_mode & S_IWOTH) != 0 )
    die(__LINE__, _("Data directory <%s> is world writeable\n"),
	DEFAULT_DATAROOT);

  if ( (buf.st_mode & S_IWGRP) != 0 && (buf.st_gid != getegid()))
    die(__LINE__, 
	_("Data directory <%s> is group writeable for another group\n"),
	DEFAULT_DATAROOT);

  if ( (buf.st_mode & S_IWGRP) != 0 )
    is_group_writeable = 1;

  /* Build full database path
   */
  o_database = xalloc (__LINE__, strlen(g_database) + 1);
  sl_strlcpy (o_database, g_database, strlen(g_database) + 1);

  str = xalloc (__LINE__, strlen(DEFAULT_DATAROOT) + strlen(g_database) + 2);
  sl_strlcpy (str, DEFAULT_DATAROOT, strlen(DEFAULT_DATAROOT) + 1);
  sl_strlcat (str, "/", strlen(DEFAULT_DATAROOT) + 2);
  sl_strlcat (str, g_database, 
	      strlen(DEFAULT_DATAROOT) + strlen(g_database) + 2);
  g_database = str;
  
  if (0 != lstat(g_database, &buf))
    die(__LINE__, _("Cannot stat() data file <%s>: %s\n"), 
        g_database, strerror(errno));

  if (!S_ISREG(buf.st_mode))
    die(__LINE__, _("Data file <%s> is not a regular file\n"),
        g_database);

  if ( (buf.st_mode & S_IWOTH) != 0 )
    die(__LINE__, _("Data file <%s> is world writeable\n"),
        g_database);

  if ( (buf.st_mode & S_IWGRP) != 0 && (buf.st_gid != getegid()))
    die(__LINE__, 
        _("Data file <%s> is group writeable for another group\n"),
        g_database);
  
  if ((buf.st_mode & S_IWGRP) != 0)
    umask (0017);
  else
    umask (0037);

  build_merge_path();

  if (g_database && g_merge && (g_update == 1))
    {
      merge_database(g_database, g_merge);
      return EXIT_SUCCESS;
    }
  else
    {
      die(__LINE__, _("No valid action specified\n"));
    }
  return EXIT_FAILURE;
}


/******************************************************************
 *
 * Command line options
 *
 ******************************************************************/
#define HAS_ARG_NO  0
#define HAS_ARG_YES 1
#define DROP_PRIV_NO  0
#define DROP_PRIV_YES 1


typedef struct options {
  char * longopt;
  const char   shortopt;
  char * usage;
  int          hasArg;
  int (*func)(char * opt);
} opttable_t;

static int sh_getopt_usage (char * dummy);
/* extern char * de_escape (char * instr); */


static int set_verbose (char * dummy);
static int set_update  (char * file);
static int set_merge   (char * file);


static opttable_t op_table[] = {
  { N_("update"),  
    'u', 
    N_("Update database"),  
    HAS_ARG_YES, 
    set_update},
  { N_("merge"),  
    'M', 
    N_("Merge file <database>.<uuid> with the file <database>"),  
    HAS_ARG_YES, 
    set_merge},
  { N_("verbose"),  
    'v', 
    N_("Be verbose"),  
    HAS_ARG_NO, 
    set_verbose },
  { N_("help"),  
    'h', 
    N_("Print usage information"),  
    HAS_ARG_NO, 
    sh_getopt_usage },
  /* last entry -- required !! -- */
  { NULL, 
    '\0',     
    NULL,  
    HAS_ARG_NO, 
    NULL }
};

static int sh_getopt_usage (char * dummy)
{
  int  i;
  char fmt[64];

  char opts[64] = { '\0' };


  fprintf (stdout,
	   _("This is %s (%s),\n"\
	     "   (c) 2015 Rainer Wichmann (http://la-samhna.de).\n"),
	   program_name, VERSION);
  fprintf (stdout, "%s", _("This software comes with ABSOLUTELY NO WARRANTY. "));
  fprintf (stdout, "%s", _("Use at own risk.\n"));

  fprintf (stdout, _("Usage: %s --merge <uuid> --update <database>\n\n"),
	   program_name);

  for (i = 0; op_table[i].shortopt != '\0'; ++i) {

    if (strchr(opts, op_table[i].shortopt) != NULL)
      fprintf (stdout, "%s", _("Short option char collision !\n"));
    opts[i] = op_table[i].shortopt;


    if (op_table[i].hasArg == HAS_ARG_NO) {
      if (strlen(op_table[i].longopt) < 10) 
	strcpy(fmt, _("-%c,        --%-s,\t\t\t %s\n")); /* known to fit  */
      else if (strlen(op_table[i].longopt) < 17)
	strcpy(fmt, _("-%c,        --%-s,\t\t %s\n"));   /* known to fit  */
      else 
	strcpy(fmt, _("-%c,        --%-s,\t %s\n"));     /* known to fit  */
      fprintf (stdout,
	       fmt,
	       op_table[i].shortopt,
	       _(op_table[i].longopt),
	       _(op_table[i].usage));
    } else {
      if (strlen(op_table[i].longopt) < 12) 
	strcpy(fmt,                                      /* known to fit  */
	       _("-%c <arg>,  --%-s=<arg>,\t\t %s\n"));  
      else 
	strcpy(fmt,                                      /* known to fit  */ 
	       _("-%c <arg>,  --%-s=<arg>,\t %s\n"));   
      fprintf (stdout,
	       fmt,
	       op_table[i].shortopt,
	       _(op_table[i].longopt),
	       _(op_table[i].usage));
    }
  }

  fprintf (stdout, "%s", 
	   _("\nPlease report bugs to support@la-samhna.de.\n"));

  (void) fflush(stdout);

  if ( dummy != NULL) 
    {
      if (strcmp( dummy, _("fail")) == 0 ) 
	{
	  fprintf(stdout, "EXIT_FAILURE 33\n");
	  (void) fflush (stdout);
	  _exit (33);
	}
    }

  fprintf(stdout, "EXIT_SUCCESS 0\n");
  (void) fflush (stdout);
  _exit (EXIT_SUCCESS);
  return 0; /* make compilers happy */
}

static int set_verbose (char * dummy)
{
  (void) dummy;
  g_verbose += 1;
  return 0;
}

static int set_update (char * str)
{
  g_database = str;
  g_update = 1;
  return 0;
}


static int set_merge (char * str)
{
  size_t len = strlen(str)+1;
  if (g_merge)
    free(g_merge);
  g_merge = xalloc(__LINE__, len);
  sl_strlcpy(g_merge, str, len);
  return 0;
}

static int sh_getopt_get (int argc, char * argv[])
{
  int           count   = 0;
  unsigned long len     = 0;
  int           foundit = 0;
  int           i;
  unsigned long k;
  char        * theequal;

  /* -- Return if no args. --
   */
  if (argc < 2) 
    return 0;
 
  while (argc > 1  && argv[1][0] == '-') 
    {

      /* Initialize
       */
      foundit = 0;
      len     = strlen (argv[1]);
    
      /* a '-' with no argument: error
       */
      if (len == 1)
	die(__LINE__, _("no argument after dash\n"));


      /* a '--' with no argument: stop argument processing
       */
      if (len == 2 && argv[1][1] == '-') 
	return count;

      /* a short option: process it
       */
      if (len >= 2 && argv[1][1] != '-') 
	{
	  for (k = 1; k < len; ++k)
	    {
	      for (i = 0; op_table[i].shortopt != '\0'; ++i) 
		{
		  
		  if ( op_table[i].shortopt == argv[1][k] ) 
		    {
		      foundit = 1;
		      if ( op_table[i].hasArg == HAS_ARG_YES ) 
			{
			  if (k != (len - 1))
			    {
			      /* not last option
			       */
			      die(__LINE__, _("short option with argument is not last in option string\n"));
			    }
			  if (argc < 3) 
			    { 
			      /* argument required, but no avail 
			       */
			      die (__LINE__, _("missing argument\n"));
			    } 
			  else 
			    {
			      /* call function with argument */
			      --argc; ++argv;
			      if (0 != (* op_table[i].func )(argv[1]))
				die (__LINE__, _("bad option: -%c %s\n"),
				     op_table[i].shortopt, argv[1]);
			      break;
			    }
			} 
		      else 
			{
			  if (0 != (* op_table[i].func )(NULL))
			    die (__LINE__, _("bad option: %s\n"), argv[1]);
			  break;
			}
		    }
		}
	    }

	  /* 'break' should get here 
	   */
	  if (foundit == 1) 
	    {
	      --argc; ++argv;
	      continue;
	    } 
	  else 
	    {
	      /* unrecognized short option */
	      die(__LINE__, _("unrecognized short option: %s\n"), argv[1]);
	    }
	}

      /* a long option: process it
       */
      if (len > 2) 
	{

	  for (i = 0; op_table[i].longopt != NULL; ++i) 
	    {
      
	      if (strncmp(_(op_table[i].longopt), 
			  &argv[1][2], 
			  strlen(op_table[i].longopt)) == 0 ) 
		{
		  foundit = 1; 
		  if ( op_table[i].hasArg == HAS_ARG_YES ) 
		    {
		      if ( (theequal = strchr(argv[1], '=')) == NULL) 
			{ 
			  if (argc < 3) 
			    { 
			      /* argument required, but no avail 
			       */
			      die(__LINE__, _("missing argument\n"));
			    } 
			  else 
			    {
			      char * larg = argv[1];
			      /* call function with argument */
			      --argc; ++argv;
			      if (NULL != op_table[i].func &&
				  0 != (* op_table[i].func )(argv[1]))
				die(__LINE__, _("bad option: %s\n"), larg);
			    }
			} 
		      else 
			{
			  if (strlen (theequal) > 1) 
			    {
			      ++theequal;
			      /* call function with argument */
			      if (0 != (* op_table[i].func )(theequal))
				die(__LINE__, _("bad option: %s\n"), argv[1]);
			      break;
			    } 
			  else 
			    {
			      die(__LINE__, _("invalid argument\n"));
			      /* argument required, but no avail */
			    }
			}
		    } 
		  else 
		    {
		      if (0 != (* op_table[i].func )(NULL))
			die(__LINE__, _("bad option: %s\n"), argv[1]);
		      break;
		    }
		}
	    }

	  /* 'break' should get here */
	  if (foundit == 1) 
	    {
	      ++count;
	      --argc; 
	      ++argv;
	      continue;
	    } 
	  else 
	    {
	      /* unrecognized long option */
	      die (__LINE__, _("unrecognized long option: %s\n"), argv[1]);
	    }
	}
    }

  return count;
}


